from .async_client import AsyncDirectInjectClient
from .client import DirectInjectClient, DirectInjectError, DirectInjectNakError
from .protocol import (
    ACK,
    ESC,
    ETX,
    NAK,
    STX,
    DiCommand,
    DirectInjectCodec,
    DiTarget,
    build_bump_sv_percent_body,
    build_param_preset_recall_body,
    build_set_string_sv_body,
    build_set_sv_body,
    build_set_sv_percent_body,
    build_subscribe_sv_body,
    build_subscribe_sv_percent_body,
    build_unsubscribe_sv_body,
    build_unsubscribe_sv_percent_body,
    build_venue_preset_recall_body,
)

__all__ = [
    "ACK",
    "ESC",
    "ETX",
    "NAK",
    "STX",
    "DiCommand",
    "DiTarget",
    "AsyncDirectInjectClient",
    "DirectInjectClient",
    "DirectInjectCodec",
    "DirectInjectError",
    "DirectInjectNakError",
    "build_bump_sv_percent_body",
    "build_param_preset_recall_body",
    "build_set_string_sv_body",
    "build_set_sv_body",
    "build_set_sv_percent_body",
    "build_subscribe_sv_body",
    "build_subscribe_sv_percent_body",
    "build_unsubscribe_sv_body",
    "build_unsubscribe_sv_percent_body",
    "build_venue_preset_recall_body",
]



from __future__ import annotations

from collections.abc import Iterable
from dataclasses import dataclass
from enum import IntEnum

STX = 0x02
ETX = 0x03
ACK = 0x06
NAK = 0x15
ESC = 0x1B

SPECIAL_BYTES = {STX, ETX, ACK, NAK, ESC}


class DiCommand(IntEnum):
    SET_SV = 0x88
    SUBSCRIBE_SV = 0x89
    UNSUBSCRIBE_SV = 0x8A
    VENUE_PRESET_RECALL = 0x8B
    PARAM_PRESET_RECALL = 0x8C
    SET_SV_PERCENT = 0x8D
    SUBSCRIBE_SV_PERCENT = 0x8E
    UNSUBSCRIBE_SV_PERCENT = 0x8F
    BUMP_SV_PERCENT = 0x90
    SET_STRING_SV = 0x91


@dataclass(frozen=True)
class DiTarget:
    node: int
    virtual_device: int
    object_id: int
    state_variable: int

    def to_bytes(self) -> bytes:
        return (
            _pack_u16(self.node)
            + _pack_u8(self.virtual_device)
            + _pack_u24(self.object_id)
            + _pack_u16(self.state_variable)
        )


class DirectInjectCodec:
    @staticmethod
    def encode(body: bytes) -> bytes:
        checksum = _checksum(body)
        escaped_body = _escape_bytes(body)
        escaped_checksum = _escape_bytes(bytes([checksum]))
        return bytes([STX]) + escaped_body + escaped_checksum + bytes([ETX])

    @staticmethod
    def decode(frame: bytes) -> bytes:
        if not frame or frame[0] != STX or frame[-1] != ETX:
            msg = "Frame must start with STX and end with ETX."
            raise ValueError(msg)
        unescaped = _unescape_bytes(frame[1:-1])
        if len(unescaped) < 2:
            msg = "Frame is too short to contain a checksum."
            raise ValueError(msg)
        body, checksum = unescaped[:-1], unescaped[-1]
        if _checksum(body) != checksum:
            msg = "Checksum does not match message body."
            raise ValueError(msg)
        return body


def build_set_sv_body(target: DiTarget, data: int) -> bytes:
    return bytes([DiCommand.SET_SV]) + target.to_bytes() + _pack_i32(data)


def build_subscribe_sv_body(target: DiTarget, rate_ms: int) -> bytes:
    return bytes([DiCommand.SUBSCRIBE_SV]) + target.to_bytes() + _pack_u32(rate_ms)


def build_unsubscribe_sv_body(target: DiTarget) -> bytes:
    return bytes([DiCommand.UNSUBSCRIBE_SV]) + target.to_bytes() + _pack_u32(0)


def build_set_sv_percent_body(target: DiTarget, percent_scaled: int) -> bytes:
    return (
        bytes([DiCommand.SET_SV_PERCENT])
        + target.to_bytes()
        + _pack_i32(percent_scaled)
    )


def build_subscribe_sv_percent_body(target: DiTarget, rate_ms: int) -> bytes:
    return (
        bytes([DiCommand.SUBSCRIBE_SV_PERCENT]) + target.to_bytes() + _pack_u32(rate_ms)
    )


def build_unsubscribe_sv_percent_body(target: DiTarget) -> bytes:
    return bytes([DiCommand.UNSUBSCRIBE_SV_PERCENT]) + target.to_bytes() + _pack_u32(0)


def build_bump_sv_percent_body(target: DiTarget, percent_scaled_delta: int) -> bytes:
    return (
        bytes([DiCommand.BUMP_SV_PERCENT])
        + target.to_bytes()
        + _pack_i32(percent_scaled_delta)
    )


def build_venue_preset_recall_body(preset_number: int) -> bytes:
    return bytes([DiCommand.VENUE_PRESET_RECALL]) + _pack_u32(preset_number)


def build_param_preset_recall_body(preset_number: int) -> bytes:
    return bytes([DiCommand.PARAM_PRESET_RECALL]) + _pack_u32(preset_number)


def build_set_string_sv_body(target: DiTarget, value: str) -> bytes:
    encoded = value.encode("ascii", errors="strict")
    if len(encoded) > 32:
        msg = "String SV values must be 32 ASCII characters or fewer."
        raise ValueError(msg)
    length = len(encoded) + 1
    data = _pack_u16(length) + encoded + b"\x00"
    return bytes([DiCommand.SET_STRING_SV]) + target.to_bytes() + data


def _escape_bytes(data: bytes) -> bytes:
    escaped = bytearray()
    for byte in data:
        if byte in SPECIAL_BYTES:
            escaped.append(ESC)
            escaped.append(byte + 0x80)
        else:
            escaped.append(byte)
    return bytes(escaped)


def _unescape_bytes(data: bytes) -> bytes:
    unescaped = bytearray()
    i = 0
    while i < len(data):
        byte = data[i]
        if byte == ESC:
            if i + 1 >= len(data):
                msg = "Escape byte at end of frame."
                raise ValueError(msg)
            unescaped.append(data[i + 1] - 0x80)
            i += 2
        else:
            unescaped.append(byte)
            i += 1
    return bytes(unescaped)


def _checksum(data: Iterable[int]) -> int:
    value = 0
    for byte in data:
        value ^= byte
    return value & 0xFF


def _pack_u8(value: int) -> bytes:
    if not 0 <= value <= 0xFF:
        msg = "Value must fit in an unsigned 8-bit field."
        raise ValueError(msg)
    return value.to_bytes(1, byteorder="big", signed=False)


def _pack_u16(value: int) -> bytes:
    if not 0 <= value <= 0xFFFF:
        msg = "Value must fit in an unsigned 16-bit field."
        raise ValueError(msg)
    return value.to_bytes(2, byteorder="big", signed=False)


def _pack_u24(value: int) -> bytes:
    if not 0 <= value <= 0xFFFFFF:
        msg = "Value must fit in an unsigned 24-bit field."
        raise ValueError(msg)
    return value.to_bytes(3, byteorder="big", signed=False)


def _pack_u32(value: int) -> bytes:
    if not 0 <= value <= 0xFFFFFFFF:
        msg = "Value must fit in an unsigned 32-bit field."
        raise ValueError(msg)
    return value.to_bytes(4, byteorder="big", signed=False)


def _pack_i32(value: int) -> bytes:
    if not -(2**31) <= value <= 2**31 - 1:
        msg = "Value must fit in a signed 32-bit field."
        raise ValueError(msg)
    return value.to_bytes(4, byteorder="big", signed=True)



from __future__ import annotations

import asyncio
import time
from dataclasses import dataclass

from .protocol import (
    ACK,
    ETX,
    NAK,
    STX,
    DirectInjectCodec,
    DiTarget,
    build_bump_sv_percent_body,
    build_param_preset_recall_body,
    build_set_string_sv_body,
    build_set_sv_body,
    build_set_sv_percent_body,
    build_subscribe_sv_body,
    build_subscribe_sv_percent_body,
    build_unsubscribe_sv_body,
    build_unsubscribe_sv_percent_body,
    build_venue_preset_recall_body,
)


class AsyncDirectInjectError(RuntimeError):
    pass


class AsyncDirectInjectNakError(AsyncDirectInjectError):
    pass


@dataclass
class AsyncDirectInjectClient:
    host: str
    port: int = 1023
    timeout: float = 1.0
    expect_ack: bool = False

    _reader: asyncio.StreamReader | None = None
    _writer: asyncio.StreamWriter | None = None

    async def connect(self) -> None:
        if self._reader is not None or self._writer is not None:
            return
        reader, writer = await asyncio.wait_for(
            asyncio.open_connection(self.host, self.port),
            timeout=self.timeout,
        )
        self._reader = reader
        self._writer = writer

    async def close(self) -> None:
        if self._writer is None:
            return
        self._writer.close()
        await self._writer.wait_closed()
        self._reader = None
        self._writer = None

    async def __aenter__(self) -> AsyncDirectInjectClient:
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc, traceback) -> None:
        await self.close()

    async def send_body(self, body: bytes, expect_ack: bool | None = None) -> bool:
        if expect_ack is None:
            expect_ack = self.expect_ack
        writer = self._require_writer()
        frame = DirectInjectCodec.encode(body)
        writer.write(frame)
        await writer.drain()
        if not expect_ack:
            return True
        return await self._read_ack()

    async def set_sv(self, target: DiTarget, data: int) -> bool:
        return await self.send_body(build_set_sv_body(target, data))

    async def subscribe_sv(self, target: DiTarget, rate_ms: int) -> bool:
        return await self.send_body(build_subscribe_sv_body(target, rate_ms))

    async def unsubscribe_sv(self, target: DiTarget) -> bool:
        return await self.send_body(build_unsubscribe_sv_body(target))

    async def set_sv_percent(self, target: DiTarget, percent_scaled: int) -> bool:
        return await self.send_body(build_set_sv_percent_body(target, percent_scaled))

    async def subscribe_sv_percent(self, target: DiTarget, rate_ms: int) -> bool:
        return await self.send_body(build_subscribe_sv_percent_body(target, rate_ms))

    async def unsubscribe_sv_percent(self, target: DiTarget) -> bool:
        return await self.send_body(build_unsubscribe_sv_percent_body(target))

    async def bump_sv_percent(
        self, target: DiTarget, percent_scaled_delta: int
    ) -> bool:
        return await self.send_body(
            build_bump_sv_percent_body(target, percent_scaled_delta)
        )

    async def venue_preset_recall(self, preset_number: int) -> bool:
        return await self.send_body(build_venue_preset_recall_body(preset_number))

    async def param_preset_recall(self, preset_number: int) -> bool:
        return await self.send_body(build_param_preset_recall_body(preset_number))

    async def set_string_sv(self, target: DiTarget, value: str) -> bool:
        return await self.send_body(build_set_string_sv_body(target, value))

    async def read_body(self) -> bytes:
        reader = self._require_reader()
        frame = await self._read_frame(reader)
        return DirectInjectCodec.decode(frame)

    def _require_reader(self) -> asyncio.StreamReader:
        if self._reader is None:
            msg = "Client is not connected."
            raise AsyncDirectInjectError(msg)
        return self._reader

    def _require_writer(self) -> asyncio.StreamWriter:
        if self._writer is None:
            msg = "Client is not connected."
            raise AsyncDirectInjectError(msg)
        return self._writer

    async def _read_ack(self) -> bool:
        reader = self._require_reader()
        deadline = time.monotonic() + self.timeout
        while time.monotonic() < deadline:
            remaining = max(0.0, deadline - time.monotonic())
            try:
                data = await asyncio.wait_for(reader.readexactly(1), timeout=remaining)
            except TimeoutError:
                break
            byte = data[0]
            if byte == ACK:
                return True
            if byte == NAK:
                msg = "Device returned NAK."
                raise AsyncDirectInjectNakError(msg)
            if byte == STX:
                frame = await self._read_frame(reader, prefix=data)
                _ = DirectInjectCodec.decode(frame)
        msg = "Timed out waiting for ACK/NAK."
        raise AsyncDirectInjectError(msg)

    async def _read_frame(
        self, reader: asyncio.StreamReader, prefix: bytes = b""
    ) -> bytes:
        buffer = bytearray(prefix)
        while True:
            if buffer and buffer[0] != STX:
                buffer.pop(0)
                continue
            chunk = await reader.readexactly(1)
            buffer.extend(chunk)
            if buffer[-1] == ETX:
                return bytes(buffer)



from __future__ import annotations

import socket
import time
from dataclasses import dataclass

from .protocol import (
    ACK,
    ETX,
    NAK,
    STX,
    DirectInjectCodec,
    DiTarget,
    build_bump_sv_percent_body,
    build_param_preset_recall_body,
    build_set_string_sv_body,
    build_set_sv_body,
    build_set_sv_percent_body,
    build_subscribe_sv_body,
    build_subscribe_sv_percent_body,
    build_unsubscribe_sv_body,
    build_unsubscribe_sv_percent_body,
    build_venue_preset_recall_body,
)


class DirectInjectError(RuntimeError):
    pass


class DirectInjectNakError(DirectInjectError):
    pass


@dataclass
class DirectInjectClient:
    host: str
    port: int = 1023
    timeout: float = 1.0
    expect_ack: bool = False

    _socket: socket.socket | None = None

    def connect(self) -> None:
        if self._socket is not None:
            return
        sock = socket.create_connection((self.host, self.port), timeout=self.timeout)
        sock.settimeout(self.timeout)
        self._socket = sock

    def close(self) -> None:
        if self._socket is None:
            return
        self._socket.close()
        self._socket = None

    def __enter__(self) -> DirectInjectClient:
        self.connect()
        return self

    def __exit__(self, exc_type, exc, traceback) -> None:
        self.close()

    def send_body(self, body: bytes, expect_ack: bool | None = None) -> bool:
        if expect_ack is None:
            expect_ack = self.expect_ack
        sock = self._require_socket()
        frame = DirectInjectCodec.encode(body)
        sock.sendall(frame)
        if not expect_ack:
            return True
        return self._read_ack()

    def set_sv(self, target: DiTarget, data: int) -> bool:
        return self.send_body(build_set_sv_body(target, data))

    def subscribe_sv(self, target: DiTarget, rate_ms: int) -> bool:
        return self.send_body(build_subscribe_sv_body(target, rate_ms))

    def unsubscribe_sv(self, target: DiTarget) -> bool:
        return self.send_body(build_unsubscribe_sv_body(target))

    def set_sv_percent(self, target: DiTarget, percent_scaled: int) -> bool:
        return self.send_body(build_set_sv_percent_body(target, percent_scaled))

    def subscribe_sv_percent(self, target: DiTarget, rate_ms: int) -> bool:
        return self.send_body(build_subscribe_sv_percent_body(target, rate_ms))

    def unsubscribe_sv_percent(self, target: DiTarget) -> bool:
        return self.send_body(build_unsubscribe_sv_percent_body(target))

    def bump_sv_percent(self, target: DiTarget, percent_scaled_delta: int) -> bool:
        return self.send_body(build_bump_sv_percent_body(target, percent_scaled_delta))

    def venue_preset_recall(self, preset_number: int) -> bool:
        return self.send_body(build_venue_preset_recall_body(preset_number))

    def param_preset_recall(self, preset_number: int) -> bool:
        return self.send_body(build_param_preset_recall_body(preset_number))

    def set_string_sv(self, target: DiTarget, value: str) -> bool:
        return self.send_body(build_set_string_sv_body(target, value))

    def read_body(self) -> bytes:
        sock = self._require_socket()
        frame = self._read_frame(sock)
        return DirectInjectCodec.decode(frame)

    def _require_socket(self) -> socket.socket:
        if self._socket is None:
            msg = "Client is not connected."
            raise DirectInjectError(msg)
        return self._socket

    def _read_ack(self) -> bool:
        sock = self._require_socket()
        deadline = time.monotonic() + self.timeout
        while time.monotonic() < deadline:
            data = sock.recv(1)
            if not data:
                continue
            byte = data[0]
            if byte == ACK:
                return True
            if byte == NAK:
                msg = "Device returned NAK."
                raise DirectInjectNakError(msg)
            if byte == STX:
                frame = self._read_frame(sock, prefix=data)
                _ = DirectInjectCodec.decode(frame)
        msg = "Timed out waiting for ACK/NAK."
        raise DirectInjectError(msg)

    def _read_frame(self, sock: socket.socket, prefix: bytes = b"") -> bytes:
        buffer = bytearray(prefix)
        while True:
            if buffer and buffer[0] != STX:
                buffer.pop(0)
                continue
            chunk = sock.recv(1)
            if not chunk:
                continue
            buffer.extend(chunk)
            if buffer[-1] == ETX:
                return bytes(buffer)
